SRCDIR = "."
import os
import sys
sys.path.append(SRCDIR)

import numpy as np
import imageio
import matplotlib.pyplot as plt
import seaborn as sns


import src.forensic_similarity as forsim #forensic similarity tool
from src.utils.blockimage import tile_image #function to tile image into blocks


def sims(img0, img1, patch_size, overlap, nb_patches):
    overlap *= patch_size
    overlap = int(overlap)
    np.random.seed(0)
    
    ROOT = os.path.dirname(os.path.realpath(__file__))
    if patch_size == 256:
        f_weights = os.path.join(ROOT, 'src/pretrained/cam_256/-30')
    elif patch_size == 128:
        f_weights = os.path.join(ROOT, 'src/pretrained/cam_128/-30')
    else:
        print('Incorrect patch size')
    
    I0 = imageio.v2.imread(img0)
    T0,xy0 = tile_image(I0,width=patch_size,height=patch_size, x_overlap=overlap,y_overlap=overlap)
    
    # COMPARE REF IMAGE TO TEST IMAGE 1
    # read test image 1
    I1 = imageio.v2.imread(img1)
    # tiles and xy coordinates of each tile for image 1
    T1,xy1 = tile_image(I1,width=patch_size,height=patch_size, x_overlap=overlap,y_overlap=overlap)
    # check that there are enough patches to do nb_patches comparisons
    if len(T0)*len(T1)>= nb_patches:
        inds_aux_0_1 = np.random.choice(range(len(T0)*len(T1)), nb_patches, replace=False) 
        inds01 = inds_aux_0_1 % len(T0) 
        inds1 = inds_aux_0_1 // len(T0) 
        X01 = T0[inds01] 
        X1 = T1[inds1]
        sim1 = forsim.calculate_forensic_similarity(X01,X1,f_weights,patch_size)
    else:
        sim1 = []
    return sim1

    
def output(img0,img1,img2, patch_size, overlap, nb_patches):
    sim1 = sims(img0, img1, patch_size, overlap, nb_patches)
    if img2 != None:
        sim2 = sims(img0, img2, patch_size, overlap, nb_patches)
    else:
        sim2 = []
        
    if np.array_equal(sim1,[]) and np.array_equal(sim2,[]):
        print(f'The images are too small to perform {nb_patches} patch-to-patch comparisons.')
    else:
        sns.set()
        plt.figure(figsize=(5, 5))
        plt.hist(sim1, bins=50, alpha=0.5, range=(0, 1), label="Reference Image - Image 1", color='#076377')
        if img2 != None:
            plt.hist(sim2, bins=50, alpha=0.5, range=(0, 1), label="Reference Image - Image 2", color='#8260a0')
        plt.xlabel("Forensic similarity score", size=9)
        plt.ylabel("Count", size=9)
        plt.title("Forensic similarity of the patch-to-patch comparisons")
        plt.legend()
        plt.savefig("output_0.png")
        plt.close()

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-input_ref", required= True)
    parser.add_argument("-input_test1", required= True)
    parser.add_argument("-input_test2", required= False)
    parser.add_argument("-patch_size", required= True)
    parser.add_argument("-overlap", required= True)
    parser.add_argument("-nb_patches", required= True)
    parser.parse_args()
    args = parser.parse_args()
    
    in1 = args.input_ref
    in2 = args.input_test1
    if os.path.exists(f'{args.input_test2}'):
        in3 = args.input_test2
    else:
        in3 = None
    patch_size = int(args.patch_size)
    overlap = int(args.overlap)
    nb_patches = int(args.nb_patches)
    output(in1,in2,in3, patch_size, overlap, nb_patches)
